@@ -139,14 +139,19 @@ def _render_ReST(
 def _render_CodeChat(text: str, filePath: str) -> Tuple[str, str]:
     try:
         rest_string = code_to_rest_string(text, filename=filePath)
-    except (KeyError, ValueError, pygments.util.ClassNotFound):
+    except KeyError as e:
+        return (
+            "",
+            f"{filePath}:: ERROR: this file, recognized as {e.args[0]}, is not supported by CodeChat.",
+        )
+    except (ValueError, pygments.util.ClassNotFound):
         # Although the file extension may be in the list of supported
         # extensions, CodeChat may not support the lexer chosen by Pygments.
         # For example, a ``.v`` file may be Verilog (supported by CodeChat)
         # or Coq (not supported). In this case, provide an error message.
         return (
             "",
-            "{}:: ERROR: this file is not supported by CodeChat.".format(filePath),
+            f"{filePath}:: ERROR: this file is not supported by CodeChat.",
         )
     return _render_ReST(rest_string, filePath, True)
 
@@ -525,11 +530,10 @@ def _select_renderer(
     bool,
 ]:
     # If this is a directory, start searching there. Otherwise, assume it's a file and remove the file name to produce a directory.
-    if not file_path.is_dir():
-        file_path = file_path.parent
+    project_path_search = file_path if file_path.is_dir() else file_path.parent
 
     # Search for an external builder configuration file. I can't find an equivalent of `parents <https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.parents>`_ that includes the full path, so the list below constructs it.
-    for project_path in [file_path, *file_path.parents]:
+    for project_path in [project_path_search, *project_path_search.parents]:
         project_file = project_path / "codechat_config.yaml"
         if project_file.exists():
             return _render_external_project, str(project_file), True
