@@ -134,6 +134,9 @@ class CodeChatHandler:
     # Indicate the server is alive.
     def ping(self) -> str:
         logger.info("ping()\n")
+        if shutdown_event.is_set():
+            # Indicate an error by returning a non-emptry string.
+            return "Shutting down."
         return ""
 
     # Render the provided text to HTML, then enqueue it for the web view.
@@ -174,12 +177,9 @@ class CodeChatHandler:
             logger.info(" => {}".format(ret))
             return ret
 
-    # _`Shut down an editor client`. The sequence is:
+    # _`Shut down an editor client`. The sequence is complex. Here's a web of links which step throughs the process.
     #
-    # #.    _`Client stop`: send a message to the web client, informing it of the shutdown. While the editor shouldn't make any more ``start_render`` or ``stop_client`` calls, doing so won't cause the server to misbehave.
-    # #.    _`Client deletion`: when the web client receives the shutdown message, tell the renderer to delete its ClientState. Since the ClientState contains the queue with the shutdown message, it shouldn't be deleted until the message is sent.
-    #
-    # This method performs the first step; ``get_result`` performs the second.
+    # _`Delete step 1.` `--> <Delete step 2.>` The editor/IDE invokes stop_client, which delegates to the render manager thread.
     def stop_client(self, id: int) -> str:
         logger.info("stop_client(id={})".format(id))
         ok = self.render_manager.threadsafe_shutdown_client(id)
@@ -218,7 +218,12 @@ def editor_plugin_server() -> None:
     # A simpler server:
     ## server = TServer.TSimpleServer(processor, transport, tfactory, pfactory)  # noqa: E266
 
-    server.serve()
+    try:
+        server.serve()
+    except Exception:
+        # Shut down the server instead of allowing it to keep running in a broken state.
+        shutdown_event.set()
+        raise
 
 
 # Server for the CodeChat Client
@@ -242,6 +247,9 @@ def client_html() -> str:
 # The endpoint for files requested by a specific client, including rendered source files. Note that ``int`` by default is `positive only <https://werkzeug.palletsprojects.com/en/2.0.x/routing/#werkzeug.routing.IntegerConverter>`_.
 @client_app.route("/client/<int(signed=True):id>/<path:url_path>")
 def client_data(id: int, url_path: str) -> Union[str, Response]:
+    # On Windows, the path begins with a drive letter; otherwise, the path begins with a ``/``, which gets absorbed into the ``/`` before the ``url_path`` component of the URL. Restore it.
+    if not render_manager.is_win:
+        url_path = "/" + url_path
     # See if we rendered this file.
     html = handler.render_manager.threadsafe_get_render_results(id, url_path)
 
@@ -307,9 +315,17 @@ def run_servers() -> int:
     )
     editor_plugin_thread.start()
 
+    def flask_launcher(*args, **kwargs):
+        try:
+            client_app.run(*args, **kwargs)
+        except Exception:
+            # Shut down the server instead of allowing it to keep running in a broken state.
+            shutdown_event.set()
+            raise
+
     # Taken from https://stackoverflow.com/a/45017691.
     flask_server_thread = threading.Thread(
-        target=client_app.run, kwargs=dict(port=HTTP_PORT), name="Flask", daemon=True
+        target=flask_launcher, kwargs=dict(port=HTTP_PORT), name="Flask", daemon=True
     )
     flask_server_thread.start()
 
